#include "global.h"
#include <omnetpp.h>
#include <list>
#include <fstream>
//#include <iostream>
//#include <sstream>

static double t_table[30][2] = {
    {6.314, 12.706},
    {2.920, 4.303},
    {2.353, 3.182},
    {2.132, 2.776},
    {2.015, 2.571},
    {1.943, 2.447},
    {1.895, 2.365},
    {1.860, 2.306},
    {1.833, 2.262},
    {1.812, 2.228},
    {1.796, 2.201},
    {1.782, 2.179},
    {1.771, 2.160},
    {1.768, 2.145},
    {1.753, 2.131},
    {1.746, 2.120},
    {1.740, 2.110},
    {1.734, 2.101},
    {1.729, 2.093},
    {1.725, 2.086},
    {1.721, 2.080},
    {1.717, 2.074},
    {1.714, 2.069},
    {1.711, 2.064},
    {1.708, 2.060},
    {1.706, 2.056},
    {1.703, 2.052},
    {1.701, 2.048},
    {1.699, 2.045},
    {1.697, 2.042}
};

struct tag_id
{
double t_in;
bool identified;
unsigned int selected_slot;
unsigned long counter;
};

class System : public cSimpleModule
{
protected:
    virtual void initialize();
public:
    Module_Class_Members(System, cSimpleModule,0);
    virtual void handleMessage(cMessage *msg);
    virtual void frameStart(cMessage *msg);
    virtual void frameStart1(cMessage *msg);
    virtual void frameStart2(cMessage *msg);
    virtual void tagIn1(cMessage *msg);
    virtual void tagIn2(cMessage *msg);
    virtual void tagOut1(cMessage *msg);
    virtual void tagOut2(cMessage *msg);
    virtual void queryResponse1(cMessage *msg);
    virtual void queryResponse2(cMessage *msg);
    virtual void endOfIdPacket1(cMessage *msg);
    virtual void endOfIdPacket2(cMessage *msg);
    virtual void loadTags1(cMessage *msg);
    virtual void loadTags2(cMessage *msg);
    virtual void nextSlot1(cMessage *msg);
    virtual void nextSlot2(cMessage *msg);
    virtual void newBatch(cMessage *msg);
    virtual void finish();
private:
    double load_area_length1;
    double load_area_length2;
    double coverage_length1;
    double coverage_length2;
    double speed1;
    double speed2;
    double max_error;
    //int hist_cells;
    unsigned int max_iterations;
    unsigned int tags_per_load1;
    unsigned int tags_per_load2;
    
    unsigned int Q;
    unsigned int N1;//First reader - Frame size
    unsigned int N2;//Second reader - Frame size
    unsigned int Q1;
    unsigned int Q2;

    unsigned int slot_counter;//Global counter
    unsigned int slot_counter1;//First reader counter
    unsigned int slot_counter2;
    unsigned int slot_counter_aux;
    unsigned int siguienteSlot1;//Counter which indicates which one's next in reading purpose
    unsigned int siguienteSlot2;
    unsigned int slotsDFSA1;//Slot number for scheduling(this is the N-scheduling parameter)
    unsigned int slotsDFSA2;
    unsigned int maximum_frame_length;
    unsigned int maximum_frame1_length;
    unsigned int maximum_frame2_length;
    unsigned int maximum_subframe1;
    unsigned int maximum_subframe2;
    unsigned int batch_count;
    unsigned long incoming_tag_counter;
    unsigned long unidentifed_tags;//Global counter
    unsigned long unidentifed_tags1;
    unsigned long unidentifed_tags2;
    unsigned long outgoing_tag_counter;//Global counter
    unsigned long outgoing_tag_counter1;
    unsigned long outgoing_tag_counter2;

    double batch_time;
    double load_time1;
    double load_time2;
    double idle_slot_time;
    double colission_slot_time;
    double successful_slot_time;
    double activation_time;
    double inter_tag_space1;
    double inter_tag_space2;
    double start_offset;
    double time_under_coverage1;
    double time_under_coverage2;

    std::list<tag_id*> load_area1;///////////// area de carga(cinta transportadora)
    std::list<tag_id*> coverage_area1;///////////// area de cobertura (alcance lector)
    std::list<unsigned int> selected_slots1; /* trama (lleva un control de los slots en los que los tags quieren transmitir su identidad) */

    std::list<tag_id*> load_area2;
    std::list<tag_id*> coverage_area2;
    std::list<unsigned int> selected_slots2;

    cStdDev global_TLR_average;
    cStdDev reader1_TLR_average;
    cStdDev reader2_TLR_average;
    unsigned int batch_number;
    std::ofstream results;
};


Define_Module(System);

void System::initialize()
{
    load_area_length1    = (double)par("load_area_length1");
    load_area_length2    = (double)par("load_area_length2");
    coverage_length1     = (double)par("coverage_length1");
    coverage_length2     = (double)par("coverage_length2");
    speed1           = (double)par("speed1");
    speed2           = (double)par("speed2");
    max_error       = (double)par("max_error");
    tags_per_load1       = (unsigned int)par("tags_per_load1");/////////////Density of each band
    tags_per_load2       = (unsigned int)par("tags_per_load2");
    Q1          = (unsigned int)par("Q1parameter");//Q in first reader
    Q2          = (unsigned int)par("Q2parameter");//Q in second reader
    Q           = Q1+Q2;//Q global

    slotsDFSA1=(unsigned int)par("slotsDFSA1");//Slot number for scheduling(this is the N-scheduling parameter)
    slotsDFSA2=(unsigned int)par("slotsDFSA2");
    siguienteSlot2=slotsDFSA1;
    siguienteSlot1=slotsDFSA2;//Set every turn

    maximum_frame_length    = (unsigned int)pow(2,Q);
    maximum_frame1_length   = (unsigned int)pow(2,Q1);
    maximum_frame2_length   = (unsigned int)pow(2,Q2);  
    inter_tag_space1     = load_area_length1 / (double)tags_per_load1;
    inter_tag_space2     = load_area_length2 / (double)tags_per_load2;
    time_under_coverage1     = coverage_length1 / speed1;/////////////
    time_under_coverage2     = coverage_length2 / speed2;/////////////
    slot_counter        = 0;
    slot_counter1        = 0;
    slot_counter2        = 0;
    slot_counter_aux        = 0;
    incoming_tag_counter    = 0;
    outgoing_tag_counter    = 0;
    outgoing_tag_counter1    = 0;
    outgoing_tag_counter2    = 0;
    
    unidentifed_tags    = 0;
    unidentifed_tags1    = 0;
    unidentifed_tags2    = 0;
    batch_count         = 0;
    max_iterations      = 40;
    batch_time      = (double)par("batch_time");
    load_time1       = load_area_length1 / speed1;
    load_time2       = load_area_length2 / speed2;
    idle_slot_time      = (double)par("idle_slot_time");
    colission_slot_time     = (double)par("colission_slot_time");
    successful_slot_time    = (double)par("successful_slot_time");
    activation_time     = (double)par("activation_time");
    start_offset        = (double)par("start_offset");
    
    
    cMessage* load_message1  = new cMessage("load message 1",LOAD_MESSAGE1);
    scheduleAt(simTime(),load_message1);
    cMessage* load_message2  = new cMessage("load message 2",LOAD_MESSAGE2);
    scheduleAt(simTime(),load_message2);
    cMessage* frame_start   = new cMessage("frame start",FRAME_START);
    scheduleAt(simTime()+start_offset+activation_time,frame_start);
    cMessage* batch     = new cMessage("batch",BATCH);
    scheduleAt(simTime()+batch_time,batch);
    
    
    const char* file    = (const char*)par("file");
    results.open(file,std::ios::app);

    load_area1.clear();
    coverage_area1.clear();
    selected_slots1.clear();

    load_area2.clear();
    coverage_area2.clear();
    selected_slots2.clear();

    //WATCH_LIST(coverage_area1);
    //WATCH_LIST(selected_slots1);
}

void System::handleMessage(cMessage *msg)
{
    switch (msg->kind()) {
        case LOAD_MESSAGE1:
            loadTags1(msg);
            break;
        case LOAD_MESSAGE2:
            loadTags2(msg);
            break;
        case FRAME_START:
            frameStart(msg);
            break;
        case FRAME_START1:
            frameStart1(msg);
            break;
        case FRAME_START2:
            frameStart2(msg);
            break;
        case TAG_IN1:
            tagIn1(msg);
            break;
        case TAG_IN2:
            tagIn2(msg);
            break;
        case TAG_OUT1:
            tagOut1(msg);
            break;
        case TAG_OUT2:
            tagOut2(msg);
            break;
        case NEXT_SLOT_1:
            nextSlot1(msg);
            break;
        case NEXT_SLOT_2:
            nextSlot2(msg);
            break;
        case QUERY_RESPONSE1:
            queryResponse1(msg);
            break;
        case QUERY_RESPONSE2:
            queryResponse2(msg);
            break;
        case END_ID_PACKET1:
            endOfIdPacket1(msg);
            break;
        case END_ID_PACKET2:
            endOfIdPacket2(msg);
            break;
        case BATCH:
            newBatch(msg);
            break;
        default:
            ev<<"ERROR: unknown message"<<endl;
            endSimulation();
    }
}

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////STEP 1
//////////////////////////////////////////////////////////////////

void System::loadTags1(cMessage *msg)
{
    for (unsigned int i = 0; i<tags_per_load1; i++)
    {
        double x    = uniform(0.0, load_area_length1);
        double t_in     = x/speed1 + simTime();
        tag_id* tag     = new tag_id;
        tag->t_in   = t_in;
        tag->identified = false;
        tag->selected_slot = 0;
        tag->counter    = incoming_tag_counter++;
        load_area1.push_back(tag);
        cMessage* tag_in = new cMessage("tag in 1",TAG_IN1);
        scheduleAt(t_in,tag_in);
    }
    scheduleAt(simTime()+load_time1,msg);   
}


void System::loadTags2(cMessage *msg)
{

  for (unsigned int i = 0; i<tags_per_load2; i++)
    {
        double x    = uniform(0.0, load_area_length2);
        double t_in     = x/speed2 + simTime();
        tag_id* tag     = new tag_id;
        tag->t_in   = t_in;
        tag->identified = false;
        tag->selected_slot = 0;
        tag->counter    = incoming_tag_counter++;
        load_area2.push_back(tag);
        cMessage* tag_in = new cMessage("tag in 2",TAG_IN2);
        scheduleAt(t_in,tag_in);
    }
    scheduleAt(simTime()+load_time2,msg);
}


void System::tagIn1(cMessage *msg)
{
    if (!load_area1.empty())
    {//extract from load area
        tag_id* tag     = load_area1.front();
        load_area1.pop_front();// From load_area to coverage_area
    //update tag info
        tag->t_in   = simTime();
    //insert in the coverage area
        coverage_area1.push_back(tag);
    //schedule tag out event
        msg->setName("tag out 1");
        msg->setKind(TAG_OUT1);
        scheduleAt(simTime()+time_under_coverage1,msg);
    }
    else
    {
        ev<<"ERROR: empty load area"<<endl;
        delete msg;
    }

}

void System::tagIn2(cMessage *msg)
{
    if (!load_area2.empty())
    {//extract from load area
        tag_id* tag     = load_area2.front();
        load_area2.pop_front();
    //update tag info
        tag->t_in   = simTime();
    //insert in the coverage area
        coverage_area2.push_back(tag);
    //schedule tag out event
        msg->setName("tag out 2");
        msg->setKind(TAG_OUT2);
        scheduleAt(simTime()+time_under_coverage2,msg);
    }
    else
    {
        ev<<"ERROR: empty load area"<<endl;
        delete msg;
    }

}


void System::tagOut1(cMessage *msg)
{
    if(!coverage_area1.empty())
    {
        outgoing_tag_counter++;
        outgoing_tag_counter1++;
        
        tag_id* tag = coverage_area1.front();
        coverage_area1.pop_front();
        if (!tag->identified)
        {
            unidentifed_tags++;
            unidentifed_tags1++; 
        }
        delete msg;
    }
    else
    {
        ev<<"ERROR: empty coverage area"<<endl;
        delete msg;
    }
}

void System::tagOut2(cMessage *msg)
{
    if(!coverage_area2.empty())
    {
        outgoing_tag_counter++;
        outgoing_tag_counter2++;

        tag_id* tag = coverage_area2.front();
        coverage_area2.pop_front();
        if (!tag->identified)
        {
            unidentifed_tags++;
            unidentifed_tags2++;
        }
        delete msg;
    }
    else
    {
        ev<<"ERROR: empty coverage area"<<endl;
        delete msg;
    }
}

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////STEP 2
//////////////////////////////////////////////////////////////////
void System::frameStart(cMessage *msg)
{
    selected_slots1.clear(); 
    selected_slots2.clear();

    slot_counter = 0;
    slot_counter1 = 0;
    slot_counter2 = 0;
    slot_counter_aux = 0;

    std::list<tag_id*>::iterator it;

    if (!coverage_area1.empty()) //if there are tags under coverage 1
    {
        for (it=coverage_area1.begin();it!=coverage_area1.end();it++)
        {
            if (!(*it)->identified)
            {//generate slot selection
                unsigned int N1      = intuniform(1,maximum_frame1_length);
                (*it)->selected_slot    = N1;
                selected_slots1.push_back(N1);
            }
        }
    }
    if (!coverage_area2.empty()) //if there are tags under coverage 2
    {
        for (it=coverage_area2.begin();it!=coverage_area2.end();it++)
        {
            if (!(*it)->identified)
            {//generate slot selection
                unsigned int N2      = intuniform(1,maximum_frame2_length);
                (*it)->selected_slot    = N2;
                selected_slots2.push_back(N2);
                }
        }
    }

////////////////////This sorts and removes duplicates in lists
    if (!selected_slots1.empty())
    {
    selected_slots1.sort();
    selected_slots1.unique();
    }
    if (!selected_slots2.empty())
    {
    selected_slots2.sort();
    selected_slots2.unique();
    }

    nextSlot1(msg);
}


void System::frameStart1(cMessage *msg)
{
    selected_slots1.clear(); //It should be empty

    slot_counter = 0;
    slot_counter1 = 0;
    //slot_counter2 = 0;
    slot_counter_aux = 0;

    std::list<tag_id*>::iterator it;

    if (!coverage_area1.empty())
    {
        for (it=coverage_area1.begin();it!=coverage_area1.end();it++)
        {
            if (!(*it)->identified)
            {//generate slot selection
                unsigned int N      = intuniform(1,maximum_frame1_length);
                (*it)->selected_slot    = N;
                selected_slots1.push_back(N);
            }
        }
    }

////////////////////This sorts and removes duplicates in lists
    if (!selected_slots1.empty())
    {
    selected_slots1.sort();
    selected_slots1.unique();
    }
    
    nextSlot1(msg);
    //std::cout<<"frame 1 constructed "<<endl;
}

void System::frameStart2(cMessage *msg)
{
    selected_slots2.clear(); //It should be empty

    slot_counter = 0;
    //slot_counter1 = 0;
    slot_counter2 = 0;
    slot_counter_aux = 0;

    std::list<tag_id*>::iterator it;

    if (!coverage_area2.empty())
    {
        for (it=coverage_area2.begin();it!=coverage_area2.end();it++)
        {
            if (!(*it)->identified)
            {//generate slot selection
                unsigned int N      = intuniform(1,maximum_frame2_length);
                (*it)->selected_slot    = N;
                selected_slots2.push_back(N);
                }
        }
    }

////////////////////This sorts and removes duplicates in lists
    if (!selected_slots2.empty()) //if there are unidentified tags under coverage 2
    {
    selected_slots2.sort();
    selected_slots2.unique();
    }
///////////////////
        
    nextSlot2(msg);
    //std::cout<<"frame constructed "<<endl;

}


void System::nextSlot1(cMessage *msg)
{
    if(siguienteSlot2==0)//Is Reader 2 turn?
    {
        siguienteSlot2=slotsDFSA1;  
        siguienteSlot1=slotsDFSA2;
        nextSlot2(msg);
    }
    else
    {//Reader 1 turn
        siguienteSlot2--;
        slot_counter1++;

        if (((double)maximum_frame1_length - slot_counter1<1.0))//Added condition of frame_lenght1  
        {
            msg->setName("frame start 1");
            msg->setKind(FRAME_START1);
            double next_frame_time  = (double)(maximum_frame1_length - slot_counter1)*idle_slot_time;
            //ev << "next_frame_time 1 = (double)(maximum_frame1_length - slot_counter1)*idle_slot_time =" << next_frame_time << endl;////////////////REMOVE
            scheduleAt(simTime()+next_frame_time+activation_time,msg);
        }
        else
        {
            msg->setName("query response 1");
            msg->setKind(QUERY_RESPONSE1);
            //ev << "Query response 1 del slot " << slot_counter1 << endl;////////////////REMOVE
            //unsigned int next_ID_slot   = selected_slots1.front();
            //ev << "next_ID_slot 1 = selected_slots1.front = " << next_ID_slot << endl;////////////////REMOVE
            //double next_response_time = (double)(next_ID_slot - slot_counter1)*idle_slot_time;
            scheduleAt(simTime()/*+next_response_time*/,msg);
        }
    }
}


void System::nextSlot2(cMessage *msg)
{
    if(siguienteSlot1==0)//Is reader 1 turn?
    {
    siguienteSlot1=slotsDFSA2;
    siguienteSlot2=slotsDFSA1;  
    nextSlot1(msg);
    }
    else
    {//Reader 2 turn
    siguienteSlot1--;
    slot_counter2++;

    if (((double)maximum_frame2_length - slot_counter2<1.0))//Added condition of frame_lenght2  
    {
        msg->setName("frame start 2");
        msg->setKind(FRAME_START2);
        double next_frame_time  = (double)(maximum_frame2_length - slot_counter2)*idle_slot_time;
        //ev << "next_frame_time 2 = (double)(maximum_frame2_length - slot_counter2)*idle_slot_time = " << next_frame_time << endl;////////////////REMOVE
        scheduleAt(simTime()+next_frame_time+activation_time,msg);
    }
    else
    {
        msg->setName("query response 2");
        msg->setKind(QUERY_RESPONSE2);
        //ev << "Query response 2 del slot " << slot_counter2 << endl;////////////////REMOVE
        //unsigned int next_ID_slot   = selected_slots2.front();
        //ev << "next_ID_slot 2 = selected_slots1.front = " << next_ID_slot << endl;////////////////REMOVE
        //double next_response_time = (double)(next_ID_slot - slot_counter2)*idle_slot_time;
        scheduleAt(simTime()/*+next_response_time*/,msg);
    }
    }
}


void System::queryResponse1(cMessage *msg)
{
    if(!selected_slots1.empty())//If there are unreads tags
    {
        int contending_tags = 0;
        slot_counter_aux = selected_slots1.front(); //update slot counter

    if(slot_counter_aux==slot_counter1)//Is this the right selected_slots1
    {
        selected_slots1.pop_front(); //delete position
    
        std::list<tag_id*>::iterator it;
        std::list<tag_id*>::iterator succesful_tag;
        for (it=coverage_area1.begin();it!=coverage_area1.end();it++)
        { //Tags with this slot
            if (((*it)->selected_slot == slot_counter1)&&(!(*it)->identified))
            {
                succesful_tag = it;
                contending_tags++;
            }
        }
        double remaining_slot_time = successful_slot_time-idle_slot_time;
        if ((contending_tags==1)&& //check if the tag remains until the end of the ID slot
        (time_under_coverage1 + (*succesful_tag)->t_in > simTime()+remaining_slot_time))
        { //successful identification slot
            (*succesful_tag)->identified = true;
            (*succesful_tag)->selected_slot = 0;
            msg->setName("end id slot");
            msg->setKind(END_ID_PACKET1);
            //ev << "successful identification slot 1"<< endl;////////////////REMOVE
            scheduleAt(simTime()+remaining_slot_time,msg);
        }
        else if (contending_tags>1)
        { //collision slot
            msg->setName("end contention slot");
            msg->setKind(END_ID_PACKET1);
            remaining_slot_time = colission_slot_time-idle_slot_time;
            //ev << "collision slot 1"<< endl;////////////////REMOVE
            scheduleAt(simTime()+remaining_slot_time,msg);
        }
        else
        { //idle slot
            msg->setName("next slot 1");
            msg->setKind(NEXT_SLOT_1);
            //ev << "idle slot 1"<< endl;////////////////REMOVE
            scheduleAt(simTime()+idle_slot_time,msg);
            //nextSlot1(msg);
        }
    }
    else
    {//This is not selected_slots1 slot
        msg->setName("next slot 1");
        msg->setKind(NEXT_SLOT_1);
        //ev << "idle slot 1 - no slot que toca"<< endl;////////////////REMOVE
        scheduleAt(simTime()+idle_slot_time,msg);   
    }
    }
    else
    {//slot with empty selected_slots
    //ev<<"ERROR: unexpected query response"<<endl;
    //delete msg;
    msg->setName("next slot 1");
    msg->setKind(NEXT_SLOT_1);
    //ev << "idle slot 1 - selected_slots esta vacia"<< endl;////////////////REMOVE
    scheduleAt(simTime()+idle_slot_time,msg);   
    }
//slot_counter1++; // Nueva forma de ir pasando slots
}



void System::queryResponse2(cMessage *msg)
{
    if(!selected_slots2.empty())
    {
        int contending_tags = 0;
        slot_counter_aux = selected_slots2.front(); //update slot counter

    if(slot_counter_aux==slot_counter2)//Is this the right selected_slots1
    {
        selected_slots2.pop_front(); //delete position

        std::list<tag_id*>::iterator it;
        std::list<tag_id*>::iterator succesful_tag;
        for (it=coverage_area2.begin();it!=coverage_area2.end();it++)
        { //Tags que tienen este slot
            if (((*it)->selected_slot == slot_counter2)&&(!(*it)->identified))
            {
                succesful_tag = it;
                contending_tags++;
            }
        }
        double remaining_slot_time = successful_slot_time-idle_slot_time;
        if ((contending_tags==1)&& //check if the tag remains until the end of the ID slot
        (time_under_coverage2 + (*succesful_tag)->t_in > simTime()+remaining_slot_time))
        { //successful identification slot
            (*succesful_tag)->identified = true;
            (*succesful_tag)->selected_slot = 0;
            msg->setName("end id slot");
            msg->setKind(END_ID_PACKET2);
            //ev << "successful identification slot 2"<< endl;////////////////REMOVE
            scheduleAt(simTime()+remaining_slot_time,msg);
        }
        else if (contending_tags>1)
        { //collision slot
            msg->setName("end contention slot");
            msg->setKind(END_ID_PACKET2);
            remaining_slot_time = colission_slot_time-idle_slot_time;
            //ev << "collision slot 2"<< endl;////////////////REMOVE
            scheduleAt(simTime()+remaining_slot_time,msg);
        }
        else
        { //idle slot
            msg->setName("next slot 2");
            msg->setKind(NEXT_SLOT_2);
            //ev << "idle slot 2"<< endl;////////////////REMOVE
            scheduleAt(simTime()+idle_slot_time,msg);
            //nextSlot2(msg);
        }
    }
    else
    {//This is not selected_slots2 slot
        msg->setName("next slot 2");
        msg->setKind(NEXT_SLOT_2);
        //ev << "idle slot 2 - no slot que toca"<< endl;////////////////REMOVE
        scheduleAt(simTime()+idle_slot_time,msg);   
    }
    }
    else
    {//slot with empty selected_slots
    //ev<<"ERROR: unexpected query response"<<endl;
    //delete msg;
    msg->setName("next slot 2");
    msg->setKind(NEXT_SLOT_2);
    //ev << "idle slot 2 - selected_slots esta vacia"<< endl;////////////////REMOVE
    scheduleAt(simTime()+idle_slot_time,msg);   
    }
//slot_counter2++; // Nueva forma de ir pasando slots
}



void System::endOfIdPacket1(cMessage *msg)
{
   nextSlot1(msg);
}

void System::endOfIdPacket2(cMessage *msg)
{
   nextSlot2(msg);
}

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////STEP 3
//////////////////////////////////////////////////////////////////


void System::newBatch(cMessage *msg)
{
    batch_count++;

    if(batch_count>2)
    {
        double TLR = (double)unidentifed_tags/(double)outgoing_tag_counter;
        double TLR1 = (double)unidentifed_tags1/(double)outgoing_tag_counter1;
        double TLR2 = (double)unidentifed_tags2/(double)outgoing_tag_counter2;
    
        global_TLR_average.collect(TLR);
        reader1_TLR_average.collect(TLR1);
        reader2_TLR_average.collect(TLR2);
        unsigned long samples = global_TLR_average.samples();
        if (samples>1)
        {
            double tlr_error = 1e6;
            double tlr_error1 = 1e6;
            double tlr_error2 = 1e6;
            if (samples<=31)
            {
                tlr_error=t_table[samples-1-1][1]*global_TLR_average.stddev()/sqrt((double)samples);
                tlr_error1=t_table[samples-1-1][1]*reader1_TLR_average.stddev()/sqrt((double)samples);
                tlr_error2=t_table[samples-1-1][1]*reader2_TLR_average.stddev()/sqrt((double)samples);
            }
            else
            {
                tlr_error=1.976*global_TLR_average.stddev()/sqrt((double)samples);
                tlr_error1=1.976*reader1_TLR_average.stddev()/sqrt((double)samples);
                tlr_error2=1.976*reader2_TLR_average.stddev()/sqrt((double)samples);
            }
            if ((tlr_error<max_error*global_TLR_average.mean())||(batch_count == max_iterations))
            {
                results<<slotsDFSA1;
                results<<"\t"<<slotsDFSA2;
                results<<"\t"<<Q1;
                results<<"\t"<<Q2;
                results<<"\t"<<global_TLR_average.mean()<<" ";
                results<<"\t"<<tlr_error<<" ";
                results<<"\t"<<reader1_TLR_average.mean()<<" ";
                results<<"\t"<<tlr_error1<<" ";
                results<<"\t"<<reader2_TLR_average.mean()<<" ";
                results<<"\t"<<tlr_error2;
                results<<"\t"<<tags_per_load1<<endl;
                endSimulation();
            }
        }
    }
    unidentifed_tags    = 0;
    unidentifed_tags1    = 0;
    unidentifed_tags2    = 0;
    outgoing_tag_counter    = 0;
    outgoing_tag_counter1    = 0;
    outgoing_tag_counter2    = 0;
    scheduleAt(simTime()+batch_time,msg);
}


void System::finish()
{
    double TLR = (double)unidentifed_tags/(double)outgoing_tag_counter;
    double TLR1 = (double)unidentifed_tags1/(double)outgoing_tag_counter1;
    double TLR2 = (double)unidentifed_tags2/(double)outgoing_tag_counter2;
    recordScalar("TLR", TLR);
    recordScalar("Perdidas_medias",global_TLR_average.mean());
    //recordScalar("Primary", TLR1);
    //recordScalar("Secondary", TLR2);
    //ev<<"maximum_frame_length = "<<maximum_frame_length<<endl;
    //ev<<"unidentified tags = "<<unidentifed_tags<<endl;
    //ev<<"outgoing_tag_counter = "<<outgoing_tag_counter<<endl;
    //ev<<"TLR = "<<TLR<<endl;

    //ev<<"total tags in "<<incoming_tag_counter<<endl;
    //ev<<"TLR "<<global_TLR_average.mean()<<endl;
    //ev<<"batches "<<batch_count<<endl;
}
